import shutil

import beautifultable
import click

from ..utils import State

ERROR_COLOR = "bright_red"
OK_COLOR = "green"
WARNING_COLOR = "bright_yellow"


def bool_text(data, false="NO", true="YES"):
    return true if data else false


def bool_text_color(data, text_false="NO", text_true="YES",
                    color_false=ERROR_COLOR,
                    color_true=OK_COLOR):
    color = color_true if data else color_false
    text = bool_text(data, text_false, text_true)
    text = click.style(text, fg=color)
    return text


def bool_text_color_inv(data, text_false="NO", text_true="YES"):
    return bool_text_color(data, text_false, text_true, OK_COLOR,
                           ERROR_COLOR)


def stop_code_text_color(code, return_msg=False):
    if code == 0:
        color = OK_COLOR
    elif code in [1, 2, 3, 4, 5]:
        color = WARNING_COLOR
    else:
        color = ERROR_COLOR

    if return_msg:
        output = '{}: {}'.format(code, State.status_meaning['stopcode'][code])
    else:
        output = '{}'.format(code)
    return click.style(output, fg=color)


def mode_text_color(motor):
    mode = motor.mode
    if mode == 'OPER':
        color = OK_COLOR
    else:
        color = ERROR_COLOR
    return click.style(mode, fg=color)


def limits_text_color(state):
    if state.is_limit_negative() and state.is_limit_positive():
        text = 'BOTH'
        color = ERROR_COLOR
    elif state.is_limit_positive():
        text = 'Lim+'
        color = WARNING_COLOR
    elif state.is_limit_negative():
        text = 'Lim-'
        color = WARNING_COLOR
    else:
        text = 'NO'
        color = OK_COLOR
    return click.style(text, fg=color)


def disable_text_color(code, return_msg=False):
    if code == 0:
        color = OK_COLOR
    elif code == 7:
        color = WARNING_COLOR
    else:
        color = ERROR_COLOR
    if return_msg:
        output = '{}: {}'.format(code,
                                 State.status_meaning['disable'][code])
    else:
        output = '{}'.format(code)
    return click.style(output, fg=color)


 

def Table(**kwargs):
    style = kwargs.pop("style", beautifultable.Style.STYLE_BOX_ROUNDED)
    kwargs.setdefault("default_padding", 1)
    kwargs.setdefault("maxwidth", shutil.get_terminal_size().columns - 1)
    kwargs.setdefault(
        "default_alignment",
        beautifultable.Alignment.ALIGN_RIGHT)
    table = beautifultable.BeautifulTable(**kwargs)
    table.set_style(style)
    return table


def StatusTable(group, style=beautifultable.Style.STYLE_BOX_ROUNDED):
    table = Table(style=style)
    table.columns.header = (
        "Axis", "Name", "Alive", "Mode", "Ready", "Power", "Disab.",
        "StopC.", "Limits")
    args = (
        group.motors,
        group.names,
        group.get_states(),
    )
    stop_codes = set()
    disable_codes = set()
    for motor, name, state in zip(*args):
        if state.get_stop_code() != 0:
            stop_codes.add(state.get_stop_code())
        if state.get_disable_code() != 0:
            disable_codes.add(state.get_disable_code())
        row = (motor.axis, name,
               bool_text_color(state.is_alive()),
               mode_text_color(motor),
               bool_text_color(state.is_ready()),
               bool_text_color(state.is_poweron(), "OFF", "ON"),
               disable_text_color(state.get_disable_code()),
               stop_code_text_color(state.get_stop_code()),
               limits_text_color(state))
        table.rows.append(row)
    output = '{}\n\n'.format(table)
    if disable_codes is not None:
        output += 'Disable codes:\n'
        for disable_code in disable_codes:
            output += '  {}\n'.format(disable_text_color(disable_code, True))
    if stop_codes is not None:
        output += 'Stop codes:\n'
        for stop_code in stop_codes:
            output += '  {}\n'.format(stop_code_text_color(stop_code, True))
    return output


def StateTable(group, style=beautifultable.Style.STYLE_BOX_ROUNDED):
    table = Table(style=style)
    table.columns.header = ('Axis', 'Name', 'Pos.', 'Ready', 'Vel.', 'Acc. T.')
    args = (group.motors, group.names, group.get_states(), group.get_fpos(),
            group.get_acctime(), group.get_velocity())
    for motor, name, state, pos, acctime, velocity in zip(*args):
        row = (motor.axis, name, pos,
               bool_text_color(state.is_ready()), velocity, acctime)
        table.rows.append(row)
    return table


def PositionTable(group, style=beautifultable.Style.STYLE_BOX_ROUNDED):
    ctrl, axes = group.controller, group.axes
    table = Table(style=style)
    header = "Axis", "AXIS", "MEASURE", "ENCIN", "INPOS", "ABSENC", "MOTOR"
    table.columns.header = header


    cols = [
        ctrl.get_pos(axes, register=register)
        for i, register in enumerate(header[1:])
    ]
    for row in zip(axes, *cols):
        table.rows.append(row)
    return table


def EncoderTable(group, style=beautifultable.Style.STYLE_BOX_ROUNDED):
    ctrl, axes = group.controller, group.axes
    table = Table(style=style)
    header = "Axis", "AXIS", "MEASURE", "ENCIN", "INPOS", "ABSENC", "MOTOR"
    table.columns.header = header
    cols = [
        ctrl.get_enc(axes, register=register)
        for i, register in enumerate(header[1:])
    ]
    for row in zip(axes, *cols):
        table.rows.append(row)
    return table


def VersionTable(group, info=True,
                 style=beautifultable.Style.STYLE_BOX_ROUNDED):
    table = Table(style=style)
    master_ver = group.controller.fver

    if not info:
        header = 'Axis', 'VER'
        for motor in group.motors:
            diff = False
            motor_ver = motor.fver
            if master_ver != motor_ver:
                diff = True

            data = [motor.addr, motor_ver]
            if diff:
                color = ERROR_COLOR
            else:
                color = OK_COLOR
            row = [click.style(str(v), fg=color) for v in data]
            table.rows.append(row)
    else:
        print('Reading...')
        header = "Axis", "SYSTEM", "DRIVER", "DSP", "FPGA"
        for motor in group.motors:
            axis_ver = motor.ver
            diff = False
            if master_ver != axis_ver.driver[0]:
                diff = True

            data = [motor.addr, axis_ver.system[0], axis_ver.driver[0],
                    axis_ver.driver_dsp[0], axis_ver.driver_fpga[0]]

            if diff:
                color = ERROR_COLOR
            else:
                color = OK_COLOR
            row = [click.style(str(v), fg=color) for v in data]

            table.rows.append(row)

    table.columns.header = header

    return table



import collections
import threading

from prompt_toolkit import HTML
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import D
from prompt_toolkit.shortcuts import ProgressBar
from prompt_toolkit.shortcuts.progress_bar import Formatter, Label, Text, \
    Percentage, Bar, Progress, TimeLeft
from prompt_toolkit.styles import Style

from ..group import gen_motion, ensure_power, gen_move, gen_rmove
from ..utils import interrupt_myself, gen_rate_limiter


class Position(Formatter):

    moving = "<moving>{}</moving>"
    stopped = "<stopped>{}</stopped>"
    pos_format = ">{width}g"

    def format(self, progress_bar, progress, width):
        state = progress.state
        template = self.moving if state.is_moving() else self.stopped
        fmt = "{{:{}}}".format(self.pos_format.format(width=width)).format
        pos = fmt(progress.position)
        return HTML(template).format(pos)

    def get_width(self, progress_bar):
        fmt = "{{:{}}}".format(self.pos_format.format(width='')).format
        lengths = [len(fmt(c.initial_position)) for c in progress_bar.counters]
        lengths += [len(fmt(c.final_position)) for c in progress_bar.counters]
        # +2 to account for - sign and space
        return D.exact(max(lengths) + 2)


class InitialPosition(Formatter):

    pos_format = ">{width}g"

    def format(self, progress_bar, progress, width):
        fmt = "{{:{}}}".format(self.pos_format.format(width=width)).format
        return fmt(progress.initial_position)

    def get_width(self, progress_bar):
        fmt = "{{:{}}}".format(self.pos_format.format(width='')).format
        lengths = (len(fmt(c.initial_position)) for c in progress_bar.counters)
        return D.exact(max(lengths))


class TargetPosition(Formatter):

    pos_format = ">{width}g"

    def format(self, progress_bar, progress, width):
        fmt = "{{:{}}}".format(self.pos_format.format(width=width)).format
        return fmt(progress.final_position)

    def get_width(self, progress_bar):
        fmt = "{{:{}}}".format(self.pos_format.format(width='')).format
        lengths = (len(fmt(c.final_position)) for c in progress_bar.counters)
        return D.exact(max(lengths))


DEFAULT_FORMATTERS = [
    Label(),
    Text(" ["),
    InitialPosition(),
    Text(" => "),
    TargetPosition(),
    Text("] "),
    Position(),
    Text(" "),
    Percentage(),
    Text(" "),
    Bar(sym_a="=", sym_b=">", sym_c=" "),
    Text(" "),
    Progress(),
    Text(" "),
    Text("eta [", style="class:time-left"),
    TimeLeft(),
    Text("]", style="class:time-left"),
    Text(" "),
]
PLAIN_FORMATTERS = [
    Label(),
    Text(" ["),
    InitialPosition(),
    Text(" => "),
    TargetPosition(),
    Text("] "),
    Position(),
    Text(" "),
    Percentage(),
]
SIMPLE_FORMATTERS = [
    Label(),
    Text(": "),
    Position(),
]
DEFAULT_STYLE = Style.from_dict({
        "moving": "DeepSkyBlue bold",
        "stopped": "bold",
        "time-left": "DarkKhaki",
        "preparing": "DarkKhaki",
        "ctrlc": "orange",
        "error": "red",
        "done": "green",
})
DEFAULT_TOOLBAR = HTML(" Press [<b>x</b>] or [<b>Ctrl-C</b>] Stop.")


def create_default_key_bindings():
    kb = KeyBindings()

    @kb.add("x")
    def _(event):
        interrupt_myself()

    return kb


Motion = collections.namedtuple(
    "Motion",
    "group start_states start_positions initial_positions final_positions")


class MotionHandler:

    def __init__(self, on_ok=None, on_error=None, always=None):
        self.on_ok = on_ok or (lambda: None)
        self.on_error = on_error or (lambda err: None)
        self.always = always or (lambda: None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_value is None:
            self.on_ok()
        else:
            self.on_error(exc_value)
        self.always()


def MotionProgressBar(motion, bar_options=None):

    def prepare_counter(counter):
        def update(state, position):
            counter.state = state
            counter.position = position
            counter.items_completed = abs(position - counter.initial_position)

        def reset(initial, final):
            counter.total = abs(final - initial)
            counter.initial_position = initial
            counter.final_position = final

        counter.update = update
        counter.reset = reset

    def update(states, positions):
        for counter, state, pos in zip(prog_bar.counters, states, positions):
            counter.update(state, pos)
        prog_bar.invalidate()

    def reset(motion):
        args = (
            prog_bar.counters,
            motion.initial_positions,
            motion.final_positions)
        for counter, initial, final in zip(*args):
            counter.reset(initial, final)
        prog_bar.invalidate()

    bar_options = bar_options or {}
    bar_options.setdefault("title", HTML("<moving>Preparing...</moving>"))
    bar_options.setdefault("formatters", DEFAULT_FORMATTERS)
    bar_options.setdefault("style", DEFAULT_STYLE)
    bar_options.setdefault("key_bindings", create_default_key_bindings())
    bar_options.setdefault("bottom_toolbar", DEFAULT_TOOLBAR)
    prog_bar = ProgressBar(**bar_options)
    prog_bar.update = update
    prog_bar.reset = reset
    names = motion.group.names
    for name, state, start, initial, final in zip(names, *motion[1:]):
        counter = prog_bar(label=name, total=abs(final - initial))
        prepare_counter(counter)
        counter.reset(initial, final)
        counter.update(state, start)
    return prog_bar


def _move(group, final_positions, refresh_period=0.1, bar_options=None):
    start_states = group.get_states()
    start_positions = group.get_pos()
    motion = Motion(
        group, start_states, start_positions, start_positions, final_positions
    )

    def motion_loop(gen):
        for states, positions in gen_rate_limiter(gen, refresh_period):
            prog_bar.update(states, positions)

    def on_ok():
        nonlocal title
        title = HTML("<done>Finished!</done>")

    def on_error(error):
        nonlocal title
        group.start_stop()
        kb = isinstance(error, KeyboardInterrupt)
        if kb:
            msg = "<ctrlc>Stopping... </ctrlc>"
        else:
            msg = "<error>Motion error: {!r}<error>. ".format(error)
            msg += "Waiting for motors to stop..."
        title = HTML(msg)
        motion_loop(gen_motion(group))
        title = HTML(msg + "[<done>DONE</done>]")

    def always_end_with():
        prog_bar.update(group.get_states(), group.get_pos())

    title = HTML("<preparing>Preparing...</preparing>")
    bar_options = bar_options or {}
    bar_options.setdefault("title", lambda: title)
    prog_bar = MotionProgressBar(motion, bar_options=bar_options)
    power = ensure_power(group)
    handler = MotionHandler(on_ok, on_error, always_end_with)
    with prog_bar, power, handler:
        title = HTML("<moving>Moving...</moving>")
        motion_loop(gen_move(group, final_positions))


def _rmove(group, deltas, refresh_period=0.1, bar_options=None):
    start_states = group.get_states()
    start_positions = group.get_pos()
    final_positions = [p + d for p, d in zip(start_positions, deltas)]
    motion = Motion(
        group, start_states, start_positions, start_positions, final_positions
    )

    def motion_loop(gen):
        for states, positions in gen_rate_limiter(gen, refresh_period):
            prog_bar.update(states, positions)

    def on_ok():
        nonlocal title
        title = HTML("<done>Finished!</done>")

    def on_error(error):
        nonlocal title
        group.start_stop()
        kb = isinstance(error, KeyboardInterrupt)
        if kb:
            msg = "<ctrlc>Stopping... </ctrlc>"
        else:
            msg = "<error>Motion error: {!r}<error>. ".format(error)
            msg += "Waiting for motors to stop..."
        title = HTML(msg)
        motion_loop(gen_motion(group))
        title = HTML(msg + "[<done>DONE</done>]")

    def always_end_with():
        prog_bar.update(group.get_states(), group.get_pos())

    title = HTML("<preparing>Preparing...</preparing>")
    bar_options = bar_options or {}
    bar_options.setdefault("title", lambda: title)
    prog_bar = MotionProgressBar(motion, bar_options=bar_options)
    power = ensure_power(group)
    handler = MotionHandler(on_ok, on_error, always_end_with)
    with prog_bar, power, handler:
        title = HTML("<moving>Moving...</moving>")
        motion_loop(gen_rmove(group, deltas))


def _rmove_multiple(group, deltas, refresh_period=0.1, bar_options=None):
    start_states = group.get_states()
    start_positions = group.get_pos()
    final_positions = [p + d for p, d in zip(start_positions, deltas)]
    dir_deltas = deltas
    rev_deltas = [-delta for delta in deltas]
    motion = Motion(
        group, start_states, start_positions, start_positions, final_positions
    )

    def motion_loop(gen):
        for states, positions in gen_rate_limiter(gen, refresh_period):
            prog_bar.update(states, positions)

    def on_ok():
        nonlocal title
        title = HTML("<done>Finished!</done>")

    def on_error(error):
        nonlocal title
        group.start_stop()
        kb = isinstance(error, KeyboardInterrupt)
        if kb:
            msg = "<ctrlc>Stopping... </ctrlc>"
        else:
            msg = "<error>Motion error: {!r}</error>. ".format(error)
            msg += "Waiting for motors to stop..."
        title = HTML(msg)
        motion_loop(gen_motion(group))
        title = HTML(msg + "[<done>DONE</done>]")

    def always_end_with():
        prog_bar.update(group.get_states(), group.get_pos())

    title = HTML("<preparing>Preparing...</preparing>")
    bar_options = bar_options or {}
    evt = threading.Event()
    keys = create_default_key_bindings()
    action = None

    @keys.add('right')
    def _(event):
        nonlocal action
        action = "right"
        evt.set()

    @keys.add('left')
    def _(event):
        nonlocal action
        action = "left"
        evt.set()
    bar_options["key_bindings"] = keys
    bar_options["bottom_toolbar"] = HTML(
        " Press [<b>x</b>] or [<b>Ctrl-C</b>] Stop | "
        '[<a bg="deepskyblue">left arrow</a>] Move left | '
        '[<a bg="deepskyblue">right arrow</a>] Move right.')
    bar_options.setdefault("title", lambda: title)
    prog_bar = MotionProgressBar(motion, bar_options=bar_options)
    power = ensure_power(group)
    handler = MotionHandler(on_ok, on_error, always_end_with)
    with prog_bar, power, handler:
        while True:
            evt.wait()
            evt.clear()
            if action == "stop":
                break
            curr_deltas = dir_deltas if action == "right" else rev_deltas
            start_pos = group.get_pos()
            final_pos = [p + d for p, d in zip(start_pos, curr_deltas)]
            motion = Motion(
                group, start_states, start_pos, start_pos, final_pos
            )
            prog_bar.reset(motion)
            action = None
            title = HTML("<moving>Moving...</moving>")
            motion_loop(gen_rmove(group, curr_deltas))
            title = HTML("<done>Finished!</done>")


def get_axes(icepap, axes_str: str):
    """
    Method to convert axes string on axes objects
    :param axes_str:
    :return:
    """
    axes_str = axes_str.strip()
    if axes_str == "all":
        axes = icepap.find_axes(only_alive=False)
    elif axes_str == "alive":
        axes = icepap.find_axes(only_alive=True)
    else:
        axes = []
        for axis in axes_str.split(","):
            try:
                axes.append(int(axis))
            except ValueError:
                axes.append(axis.strip())
        axes.sort()
    return icepap[axes]


import click
import beautifultable


from .progress_bar import DEFAULT_FORMATTERS, PLAIN_FORMATTERS, \
    SIMPLE_FORMATTERS, _move, _rmove, _rmove_multiple
from .tables import Table, StateTable, StatusTable, PositionTable, \
    VersionTable, EncoderTable
from ..group import Group
from ..controller import IcePAPController
from .utils import get_axes


class ProgressBarFormats(click.Choice):

    formatters = dict(
        default=DEFAULT_FORMATTERS,
        plain=PLAIN_FORMATTERS,
        simple=SIMPLE_FORMATTERS
    )

    def __init__(self):
        super().__init__(self.formatters)

    def convert(self, value, param, ctx):
        result = super().convert(value, param, ctx)
        return self.formatters[result]


class TableStyles(click.Choice):

    name = "table_style"

    def __init__(self):
        styles = [s.name.lstrip("STYLE_").lower()
                  for s in beautifultable.Style]
        super().__init__(styles, case_sensitive=False)

    def convert(self, value, param, ctx):
        result = super().convert(value, param, ctx)
        return beautifultable.Style["STYLE_" + result.upper()]


class Racks(click.ParamType):
    name = 'racks'

    def convert(self, value, param, ctx):
        ipap = ctx.obj["icepap"]
        if value == 'all':
            racks = ipap.find_racks()
        else:
            racks = [int(v) for v in value.split(",")]
        return racks
    

def cli_move(group, positions, format=None, bottom_toolbar=True, title=True):
    bar_options = dict(formatters=format)
    if not bottom_toolbar:
        bar_options["bottom_toolbar"] = None
    if title is not True:
        bar_options["title"] = None if title is False else title
    _move(group, positions, bar_options=bar_options)


def cli_rmove(
        group,
        deltas,
        format=None,
        bottom_toolbar=True,
        title=True,
        multiple=False):
    bar_options = dict(formatters=format)
    if not bottom_toolbar:
        bar_options["bottom_toolbar"] = None
    if title is not True:
        bar_options["title"] = None if title is False else title
    if multiple:
        _rmove_multiple(group, deltas, bar_options=bar_options)
    else:
        _rmove(group, deltas, bar_options=bar_options)


opt_mandatory_rack = click.option(
    "--rack", type=int, default=None, show_default=True,
    help="rack number. Default is None, meaning the whole system"
)

opt_mandatory_racks = click.option(
    "--racks", type=Racks(), required=True, show_default=True,
    help="comma separated list of racks. Also supports 'all'"
)

opt_racks = click.option(
    "--racks", "racks", type=Racks(), default="all", show_default=True,
    help="comma separated list of racks. Also supports 'all'"
)


@click.group(invoke_without_command=True)
@click.pass_context
@click.argument("icepap", type=IcePAPController.from_url)
@click.option("--axes", "axes_str", type=str, default="alive",
              show_default=True,
              help="comma separated list of axes. Also supports 'all' and "
                   "'alive'")
@click.option("--table-style", type=TableStyles(), default="compact",
              show_default=True, help="table style")
@click.option("--pb-format", type=ProgressBarFormats(), default="default",
              show_default=True, help="progress bar style")
@click.option("--title/--no-title", default=True, show_default=True,
              help="show/hide title")
@click.option("--bottom-toolbar/--no-bottom-toolbar", default=True,
              show_default=True, help="show/hide bottom toolbar")
def cli(ctx, icepap, axes_str, table_style, pb_format, title, bottom_toolbar):
    """
    High level Command Line Interface for IcePAP

    Connects to the given ICEPAP (a url in format [tcp://]<host/ip>[:<port=5000])
    (ex: 'ice1', 'tcp://ice1' and 'tcp://ice1:5000' all mean the same)

    Runs without command will open the Console Application
    (ex: icpapctl <url>)
    """
    ctx.ensure_object(dict)
    ctx.obj["icepap"] = icepap
    ctx.obj["axes_str"] = axes_str
    ctx.obj['axes'] = get_axes(icepap, axes_str)
    alive_axes = get_axes(icepap, 'alive')
    all_axes = get_axes(icepap, 'all')
    not_alive_axes = list(set(all_axes) - set(alive_axes))
    if axes_str == 'all':
        click.echo('Warning: There are not alive axes: {}'.format(
            ', '.join(not_alive_axes)))
    ctx.obj['table_style'] = table_style
    ctx.obj['pb_format'] = pb_format
    ctx.obj['title'] = title
    ctx.obj['bottom_toolbar'] = bottom_toolbar
    if ctx.invoked_subcommand is None:
        from .repl import run
        run(ctx)


@cli.command()
@click.pass_context
@click.argument("pairs", nargs=-1, type=str, required=True)
def mv(ctx, pairs):
    """
    Move specified axes in pairs <axis, position> to the specified absolute
    positions. If the motor is off, the command will turn ON and at the end
    will turn OFF again
    """
    ipap = ctx.obj["icepap"]
    pb_format = ctx.obj['pb_format']
    title = ctx.obj['title']
    bottom_toolbar = ctx.obj['bottom_toolbar']
    motors = [ipap[int(address)] for address in pairs[::2]]
    positions = [int(position) for position in pairs[1::2]]
    cli_move(Group(motors), positions, pb_format, bottom_toolbar, title)


@cli.command()
@click.pass_context
@click.argument("pairs", nargs=-1, type=str, required=True)
@click.option("-m", "--multiple", is_flag=True)
def mvr(ctx, pairs, multiple):
    """
    Move specified axes in pairs <axis, position> relative to their current
    positions.If the motor is off, the command will turn ON and at the end
    will turn OFF again
    """
    ipap = ctx.obj["icepap"]
    pb_format = ctx.obj['pb_format']
    bottom_toolbar = ctx.obj['bottom_toolbar']
    title = ctx.obj['title']
    motors = [ipap[int(address)] for address in pairs[::2]]
    deltas = [int(position) for position in pairs[1::2]]
    cli_rmove(
        Group(motors),
        deltas,
        pb_format,
        bottom_toolbar,
        title,
        multiple)


@cli.command()
@click.pass_context
def state(ctx):
    """Prints a summary of each axis state in form of table"""
    motors = ctx.obj['axes']
    table_style = ctx.obj['table_style']
    group = Group(motors)
    click.echo(StateTable(group, style=table_style))


@cli.command()
@click.pass_context
def status(ctx):
    """
    Prints a summary of each axis status (position velotiy, acc. time, etc)
    in form of table
    """
    motors = ctx.obj['axes']
    table_style = ctx.obj['table_style']
    group = Group(motors)
    click.echo(StatusTable(group, style=table_style))


@cli.command()
@click.pass_context
@click.option('--enc', is_flag=True, default=False,
              help='Get encoders registers')
def wa(ctx, enc):
    """Prints a summary of each axis detailed position in form of table"""
    motors = ctx.obj['axes']
    table_style = ctx.obj['table_style']
    group = Group(motors)
    if not enc:
        click.echo('Unit: Axis Steps')
        click.echo(PositionTable(group, style=table_style))
    else:
        click.echo('Unit: Encoder Steps')
        click.echo(EncoderTable(group, style=table_style))


@cli.command()
@click.pass_context
@click.option('-v', 'verbose', is_flag=True, default=False,
              help='Get all info')
@click.option('--saved', is_flag=True, default=False, help='Get saved info')
@click.option('--axes', is_flag=True, default=False, help='Get axes version')
def version(ctx, verbose, saved, axes):
    """Prints a summary of icepap version"""
    ipap = ctx.obj["icepap"]
    if axes and saved:
        click.echo('Only one option is valid --axes or --saved', color='red')
        return
    if axes:
        motors = ctx.obj['axes']
        table_style = ctx.obj['table_style']
        group = Group(motors)
        click.echo(VersionTable(group, verbose, style=table_style))
        return

    if not saved:
        click.echo(ipap.ver)
    else:
        click.echo(ipap.ver_saved)


@cli.command()
@click.pass_context
@opt_mandatory_rack
@click.confirmation_option(
    prompt="Are you sure you want to reset the rack(s)?"
)
def reset(ctx, rack):
    """Resets the given rack. If no rack is given, resets the whole system"""
    ipap = ctx.obj["icepap"]
    ipap.reset(rack)


@cli.command()
@click.pass_context
@click.confirmation_option(
    prompt="Are you sure you want to reboot?"
)
def reboot(ctx):
    """Reboots the icepap"""
    ipap = ctx.obj["icepap"]
    ipap.reboot()


@cli.command()
@click.pass_context
@opt_racks
def rackinfo(ctx, racks):
    """Prints a summary of each rack information in form of table"""
    ipap = ctx.obj["icepap"]
    rids = ipap.get_rid(racks)
    temps = ipap.get_rtemp(racks)
    table_style = ctx.obj['table_style']
    table = Table(style=table_style)
    table.columns.header = ("Rack #", "RID", "Temp.")
    for row in zip(racks, rids, temps):
        table.rows.append(row)
    click.echo(table)


@cli.command()
@click.pass_context
@click.argument("cmd", nargs=-1, type=str, required=True)
def send(ctx, cmd):
    """Send raw command to the icepap"""
    ipap = ctx.obj['icepap']
    cmd = ' '.join(cmd)
    output = ipap.send_cmd(cmd)
    if output is not None:
        # Allow name with spaces:
        if not ipap.multiline_answer:
            click.echo(' '.join(output))
        else:
            for line in output:
                click.echo(line)
    else:
        click.echo('Done')


@cli.command()
@click.pass_context
@click.argument("cmd", type=str, required=True)
def sendall(ctx, cmd):
    """ Send raw command to selected axes with option --axes """
    axes = ctx.obj['axes']
    for axis in axes:
        try:
            output = axis.send_cmd(cmd)
            if output is not None:
                click.echo('Axis {} anwser:'.format(axis.axis))
                for line in output:
                    click.echo('  {}'.format(line))
                click.echo('-'*40)
        except Exception as e:
            click.echo('Error sending command to {}'.format(axis.axis),
                       color='red')
            click.echo(e, color='red')
            click.echo('-' * 20)


if __name__ == "__main__":
    cli()



import pytest

from icepap.group import Group, group


def test_group(smart_pap):
    m1, m5 = smart_pap[1, 5]
    grp = Group(smart_pap[1, 5])

    assert grp.controller == m1._ctrl
    assert grp.motors == [m1, m5]
    assert grp.names == ['th', 'tth']

    assert grp.get_pos() == [55, -3]
    assert grp.get_fpos() == [55, -3]

    expected_states = [m1.state.status_register, m5.state.status_register]
    assert [s.status_register for s in grp.get_states()] == expected_states

    assert grp.is_moving() is False

    grp = Group(smart_pap[1])
    assert grp.motors == [m1]

    grp = Group(smart_pap[1, 151])
    assert grp.get_power() == [True, False]
    assert grp.get_acctime() == [0.1, 0.25]
    assert grp.get_velocity() == [100, 1002]


def test_group_creation_helper(smart_pap):
    m1, m5 = smart_pap[1, 5]
    grp1 = group(*smart_pap[1, 5])
    assert grp1.motors == [m1, m5]
    assert grp1.names == ['th', 'tth']
    assert group(grp1) is grp1

    grp2 = group(smart_pap[1, 5])
    assert grp1.motors == grp2.motors

    m151, m152, m153 = smart_pap[151, 152, 153]
    grp3 = group(smart_pap[151, 152])
    grp4 = group(grp1, m153, grp3)
    assert grp4.motors == [m1, m5, m153, m151, m152]
